import socket
import signal
import pyuv
import json
from time import *

g_listenip = "0.0.0.0"
g_multicast_port = 30001
g_multicast_addr = "239.255.0.1"

JsonType = 0
ProtoType = 1
topicType = 0
serviceType = 1

# json callback info
class JsnCallBack(object):
    def __init__(self):
        topicName = ""
        jsnSub = None     # std::function<void(const std::string & topic, const JsnCb &cb)> 
        jsnCb = None    # std::function<void (google::protobuf::Message *msg)> 

class MsgCallBack(object):
    def __init__(self):
        topicName = ""
        msgSub = None      # std::function<void(const std::string & topic, const MsgCb &cb)>
        msgCb = None        

class ServiceCall(object):
    def __init__(self):
        serviceName = ""
        servCall = ""   # std::function<Json(const std::string & service, Json& params ,std::function<void(Json json)> &callback)>
        callback  = None  # std::function<void(Json json)>
        params = {}     # json params

class Multicast(object):

    # multicast init
    def __init__(self,multicast_addr,localip):
        # get local socket name
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.connect(("8.8.8.8",80))
        self.__m_localip = sock.getsockname()[0]
        # print("local ip ",self.__m_localip)

        # init recv_socket
        self.loop = pyuv.Loop.default_loop()
        self.__m_multicast_addr = multicast_addr
        self.__m_recv_socket = pyuv.UDP(self.loop)
        self.__m_recv_socket.bind((g_listenip,g_multicast_port),pyuv.UV_UDP_REUSEADDR)
        self.__m_recv_socket.set_membership(multicast_addr,pyuv.UV_JOIN_GROUP)
        self.__m_recv_socket.set_broadcast(True)
        self.__m_recv_socket.start_recv(self.handlerRecv)   # recieve handler

        # init send_socket
        self.__m_send_socket = pyuv.UDP(self.loop)
        self.__m_send_socket.bind((g_listenip,g_multicast_port),pyuv.UV_UDP_REUSEADDR)
        # self.__m_send_socket.set_membership(multicast_addr,pyuv.UV_JOIN_GROUP)

        # 
        self.__m_localEndpoint = ""
        self.__m_localrpcEndpoint = ""
        self.__m_jsnCbMap = {}
        self.__m_msgCbMap = {}
        self.__m_serviceCallMap = {}
        self.__m_topicMap = {}  # std::unordered_map<std::string, std::unordered_set<std::string>>
        self.__m_serviceMap = {} # std::unordered_map<std::string, std::unordered_set<std::string>>

        self.__m_signal_h = pyuv.Signal(self.loop)
        self.__m_signal_h.start(self.signal_handler, signal.SIGINT)
    
    def get_localip(self):
        return self.__m_localip

    # if ctrl + c signal , then close socket
    def signal_handler(self,handle, signum):
        self.__m_signal_h.close()
        self.__m_recv_socket.close()
        self.__m_send_socket.close()

    def __del__( self ):
        # deconstruct function
        pass
    
    def setLocalEndpoint(self,pubsubEndpoing,rpcEndpoint):
        self.__m_localEndpoint = pubsubEndpoing
        self.__m_localrpcEndpoint = rpcEndpoint
    

    def dealDelayCallBack(self,topicName):
        # jsoncall = self.__m_jsnCbMap[topicName]
        jsoncall = self.__m_jsnCbMap.get(topicName)
        if(jsoncall != None):
            jsncb = jsoncall.jsnCb
            jsoncall.jsnSub(topicName,jsncb)   # call json sub
            del self.__m_jsnCbMap[topicName]
        # call msg sub for protobuf
        msgcall = self.__m_msgCbMap.get(topicName)
        if(msgcall != None):
            msgcb = msgcall.msgCb
            msgcall.msgSub(topicName,msgcb)   # call json sub
            del self.__m_msgCbMap[topicName]

    def dealServiceDelayCall(self,serviceName):
        # service_call = self.__m_serviceCallMap[serviceName]
        service_call = self.__m_serviceCallMap.get(serviceName)
        if(service_call != None):
            params = service_call.params
            callback = service_call.callback
            service_call.servCall(serviceName,params,callback)
            del self.__m_serviceCallMap[serviceName]

    # update topic map or service map
    def updateMap(self,type,result):
        if(type == topicType):
            print("update topic map")
            # endpoints = self.__m_topicMap[result["topic"]]   #topic may has sevel server endpoints
            endpoints = self.__m_topicMap.get(result["topic"])   #topic may has sevel server endpoints
            endpoint = result["endpoint"]
            if(endpoints != None):  # find topic and related endpoints list
                if(endpoint in endpoints):   # alread has endpoint in list
                    return
                else:
                    self.__m_topicMap[result["topic"]].append(endpoint)
            else:   # not found then insert
                ends = []
                ends.append(endpoint)
                self.__m_topicMap[result["topic"]] = ends

        # service type
        elif(type == serviceType):   
            print("update service map")
            # endpoints = self.__m_serviceMap[result["service"]]
            endpoints = self.__m_serviceMap.get(result["service"])
            endpoint = result["endpoint"]
            if(endpoints != None):  # find topic and related endpoints list
                if(endpoint in endpoints):   # alread has endpoint in list
                    return
                else:
                    self.__m_serviceMap[result["service"]].append(endpoint)
            else:   # not found then insert
                ends = []
                ends.append(endpoint)
                self.__m_serviceMap[result["service"]] = ends

    def findLocalSupport(self,supportName,type):
        if(type == topicType):
            self.sendRspToPeer(self.__m_topicMap,supportName,self.__m_localEndpoint, "topic")
        elif(type == serviceType):
            self.sendRspToPeer(self.__m_serviceMap,supportName,self.__m_localrpcEndpoint, "service")

    # map : {service_name/topic_name,[endpoints]}
    def sendRspToPeer(self,map,supportName,endpoint,field):
        # print("start to send response")
        map.setdefault(supportName,None)    # set default to prevent code collapse
        endpoints = map[supportName]
        if(endpoints != None):
            if endpoint in endpoints:
                json_dict = {
                    "method" : "response",
                    field : supportName,
                    "endpoint" : endpoint
                }
                json_str = json.dumps(json_dict)
                json_bytes = bytes(json_str, encoding = "utf8")
                self.__m_recv_socket.send((self.__m_multicast_addr,g_multicast_port),json_bytes)

    def multicastTopic(self,topicName,endpoint):
        json_dict = {
                    "method" : "commonMulti",
                    "topic" : topicName,
                    "endpoint" : endpoint
                }
        json_str = json.dumps(json_dict)
        json_bytes = bytes(json_str, encoding = "utf8")
        # print("__m_multicast_addr is ",self.__m_multicast_addr)
        # print("g_multicast_port is ",g_multicast_port)
        self.__m_send_socket.send((self.__m_multicast_addr,g_multicast_port),json_bytes)

    def sendRequest(self,supportName,filed):
        json_dict = {
                    "method" : "request",
                    filed : supportName,
                    "endpoint" : ""
                }
        # print("send request") 
        json_str = json.dumps(json_dict)
        json_bytes = bytes(json_str, encoding = "utf8")
        self.__m_send_socket.send((self.__m_multicast_addr,g_multicast_port),json_bytes)
    
    def multicastService(self,serviceName,endpoint):
        json_dict = {
                    "method" : "commonMulti",
                    "service" : serviceName,
                    "endpoint" : endpoint
                }
        # print("send service endpoint")
        json_str = json.dumps(json_dict)
        json_bytes = bytes(json_str, encoding = "utf8")
        print("multicastService : ",json_str)
        self.__m_send_socket.send((self.__m_multicast_addr,g_multicast_port),json_bytes)

    # endpoints : endpoint list
    def findTopic(self,topicName,endpoints = []):
        res = True
        # print("finding topic name in local map: " ,topicName)
        topic_endpoints = self.__m_topicMap.get(topicName)
        if(topic_endpoints == None): #can not find topic
            # print("can not find topic name in local map")
            res = False
            return res

        for endpoint in topic_endpoints:
            endpoints.append(endpoint)
        # print("findTopic : endpoints is ",endpoints)
        return res

    # endpoints : endpoint list
    def findService(self,serviceName,endpoints):
        res = True
        # print("finding service name in local map: " ,serviceName)
        # service_endpoints = self.__m_serviceMap[serviceName]
        service_endpoints = self.__m_serviceMap.get(serviceName)
        if(service_endpoints == None): #can not find topic
            res = False
            return res

        for endpoint in service_endpoints:
            endpoints.append(endpoint)
        return res

    # std::unordered_map<std::string, JsnCallBack>
    #     struct JsnCallBack{
    #     std::string     topicName;
    #     std::function<void(const std::string & topic, const JsnCb &cb)> jsnSub;
    #     JsnCb           jsnCb;
    #     ~JsnCallBack(){}
    # };
    # JsnCallBack
    def pushTopic_JsonCallBack(self,JsnCallBack):
        self.__m_jsnCbMap[JsnCallBack.topicName] = JsnCallBack
    
    # protobuf msg callback
    def pushTopic_MsgCallBack(self,MsgCallBack):
        self.__m_msgCbMap[MsgCallBack.topicName] = MsgCallBack
    
    def pushServiceCallBack(self,serviceCall):
        self.__m_serviceCallMap[serviceCall.serviceName] = serviceCall
    
    def run(self):
        # print("running multicast node")
        self.loop.run()     #run multicast service
        # print("socket disconnect")
    
    def handlerRecv(self,handle, ip_port, flags, data, error):
        # print("get into handler Recv")
        if(error is None):
            if(data is not None):
                json_str = str(data, encoding = "utf-8")  
                result = json.loads(json_str)
                # print("recieve json is : ",result)

                if("endpoint" in result.keys()):
                    method = result["method"]
                    # if recieve topic msg
                    if("topic" in result.keys()):
                        if(method == "commonMulti"):
                            self.updateMap(topicType,result)
                            self.dealDelayCallBack(result["topic"])
                        elif(method == "request"):
                            self.findLocalSupport(result["topic"],topicType)
                        elif (method == "response"):
                            self.updateMap(topicType,result)
                            self.dealDelayCallBack(result["topic"])
                    # deal service msg
                    elif("service" in result.keys()):
                        if(method == "commonMulti"):
                            self.updateMap(serviceType,result)
                            self.dealServiceDelayCall(result["service"])
                        elif(method == "request"):
                            self.findLocalSupport(result["service"],serviceType)
                        elif(method == "response"):
                            self.updateMap(serviceType,result)
                            self.dealServiceDelayCall(result["service"])
        else:
            print("error is ",error)

# def multicast_test2(mult2):
#     mult2.run()

# def multicast_test1(mult1):
#     # mult1.multicastService("add","192.168.1.3")
#     # mult1.multicastTopic("sub","192.34.2.2")
#     mult1.sendRequest("topic_request","topic")
#     # mapset = {"topic_response" : ["192","168"]}
#     # mult1.sendRspToPeer(mapset,"topic_response","192","topic")

# import threading

# if __name__ == "__main__":
#     mult1 = Multicast("239.255.0.1",10001)
#     mult2 = Multicast("239.255.0.1",10006)
#     thread1 = threading.Thread(target=multicast_test1,args=(mult1,),name="thread1")
#     thread2 = threading.Thread(target=multicast_test2,args=(mult2,),name="thread2")

#     thread1.start()
#     thread2.start()

#     mult1.run()
#     thread1.join()
#     thread2.join()



import json
import zmq
import threading
from threading import Lock
from concurrent.futures import ThreadPoolExecutor
import time

class ZMQRPCServer(object):
    m_instance = None
    m_mutex = Lock()

    def __init__(self):
        self.m_num = 0
        self.m_pool = None
        self.m_servers = None
        self.m_workers = None
        self.m_context = None
        self.m_callback = {"sub":None}
        pass

    # singleton
    @classmethod
    def getInstance(cls):
        if(cls.m_instance == None):
            cls.m_mutex.acquire()
            if(cls.m_instance == None):
                cls.m_instance = ZMQRPCServer()
            cls.m_mutex.release()
        return cls.m_instance

    def initServer(self,endpoint,threadCount):
        self.m_num = threadCount
        self.m_pool = ThreadPoolExecutor(max_workers=threadCount)
        # print("ZMQ RPC Binding to ",endpoint)
        self.m_context = zmq.Context()
        self.m_workers = self.m_context.socket(zmq.DEALER)
        self.m_workers.bind("inproc://workers")

        self.m_servers = self.m_context.socket(zmq.ROUTER)
        self.m_servers.bind(endpoint)
    
    def registerCall(self,name,callback):
        self.m_callback[name] = callback    
        # print("m_callback is ",self.m_callback)
    
    def send_str(self,socket,response):
        json_bytes = bytes(response, encoding = "utf8") #json to bytes
        socket.send(json_bytes)
    
    def DealRequest(self):
        context = ZMQRPCServer.getInstance().m_context
        socket = context.socket(zmq.REP)
        socket.connect("inproc://workers")  #connect proxy workers
        while(1):
            try:
                request = socket.recv()
                # print("recieve request is ",request)
                json_str = str(request, encoding = "utf-8")  
                json_dict = json.loads(json_str)
                if(json_dict.get("params") == None):
                    json_dict["params"] = None
                try:
                    if( "jsonrpc" in json_dict and "method" in json_dict and "id" in json_dict):
                        if( json_dict["jsonrpc"] != "2.0"):
                            raise Exception("Invalid Jsonrpc version, version must be 2.0")
                        # print("method is ",json_dict["method"])
                        # print("m_num is ",self.m_num)
                        # print("m_callback is ",self.m_callback) 
                        if( json_dict["method"] in self.m_callback.keys()):  # method found in register
                            callback = self.m_callback[json_dict["method"]]
                            response_result = callback(json_dict["id"],json_dict["params"])
                            self.send_str(socket,response_result)

                except Exception as e:
                    print("jsonrpc server excetion ",e)

            except zmq.error as e:
                print("zmq exception ({0}): {1}".format(e.errno, e.strerror))
                sys.exit()

    def run(self):
        for index in range(self.m_num):
            # print("get into run")
            self.m_pool.submit(self.DealRequest)
        zmq.proxy(self.m_servers,self.m_workers)




import zmq
import json
import threading
from threading import Lock

class ZMQRPCClient(object):
    def __init__(self):
        self.m_context = zmq.Context()
        #  Socket to talk to server
        self.m_socket = self.m_context.socket(zmq.REQ)
        self.m_lock = Lock()
    
    def connect(self,endpoint):
        self.m_socket.connect(endpoint)
    
    def make_json_request(self,id,method,params = None):
        json_dict = {
		"jsonrpc" : "2.0",
		"method": method,
		"id": id
        }

        if(params != None):
            json_dict["params"] = params

        json_str = json.dumps(json_dict)
        json_bytes = bytes(json_str, encoding = "utf8")
        return json_bytes

    # string method , json params , return json
    def call(self, method, params =None):
        data = self.make_json_request(1,method,params)

        self.m_lock.acquire()
        self.m_socket.send(data)

        reply = self.m_socket.recv()
        self.m_lock.release()
        # change bytes to json and return
        # print("client recv data is ",reply)
        json_str = str(reply, encoding = "utf-8")  
        result = json.loads(json_str)
        # print("recieve json is : ",result)

        return result
    
    def async_call(self,method,params,callback):
        if(callable(callback) != True):
            print("callback must be function")
            return -1

        new_thread = threading.Thread(target=self.async_call_proc,args=(method,params,callback))
        new_thread.start()
    
    def async_call_proc(self,method,params,callback):
        result = self.call(method,params)
        callback(result)



import json
import zmq
import threading
from threading import Lock
from concurrent.futures import ThreadPoolExecutor
import time
from Baos.multicast import *
from Baos.ZMQRPCServer import *
from Baos.ZMQRPCClient import *

import asyncio
import socket

# socket info
class socketInfo(object):
    def __init__(self,socket,flag):
        self.socket = socket
        self.flag = flag

# auto auto lock class 
# input a lock and start lock when init
# When out of scope, the destructor is called to release the lock
class AutoLock:
    def __init__(self,lock):
        self._lock = lock
        self._lock.acquire()

    def __del__(self):
        self._lock.release()

class BaosNodeImpl(object):
    m_instance = None
    m_poolSize = 10

    def __init__(self):
        self.m_mutex = []
        self.m_index = 0
        for i in range(BaosNodeImpl.m_poolSize):
            self.m_mutex.append(Lock())
        self.m_connManager = None
        self.m_pool = None
        self.m_servers = None
        self.m_workers = None
        self.m_context = None
        self.m_callback = {"sub":None}
        self.m_jsonHandlers = {}    #{string,[jscb]}
        self.m_msgHandlers = {}     #{string,[msgcb]}
        self.m_msgQueue = [i for i in range(BaosNodeImpl.m_poolSize)]    # message queue to save recieve
        self.m_connection = {}
        self.m_connectionDealWay = {}   # {string,[int]}
        self.m_connectionJson = {}  # {string,socket}

        self.m_localIP = None
        self.m_pubsubPort = None
        self.m_rpcPort = None


        self.m_poller = zmq.Poller() 
        self.m_socketInsVec = [] # [{flag,socket}]
        self.m_websocket = None

        pass
    
    def initialize(self, multiIP, point = None):
        self.m_connManager = Multicast(multiIP,self.m_localIP)
        self.m_localIP = self.m_connManager.get_localip()
        print("self.m_localIP is ",self.m_localIP)
        
        # get free port
        self.m_rpcPort = str(self.get_free_port())
        self.m_pubsubPort = str(self.get_free_port())

        # pub&sub init
        self.m_topicEndpoint = "tcp://" + self.m_localIP + ":" + self.m_pubsubPort
        print("ZMQ PUB Binding to ",self.m_topicEndpoint)
        self.m_context = zmq.Context()
        self.m_pub = self.m_context.socket(zmq.PUB)
        self.m_pub.bind(self.m_topicEndpoint)

        # zmq rpc server init
        self.m_serviceEndpoint =  "tcp://" + self.m_localIP + ":"  + self.m_rpcPort
        ZMQRPCServer.getInstance().initServer(self.m_serviceEndpoint, BaosNodeImpl.m_poolSize)

        self.m_rpcConnection = {}

        # multicast run and basnodeimpl's dealpoller run
        self.m_connManager.setLocalEndpoint(self.m_topicEndpoint, self.m_serviceEndpoint)
        self.m_pool = ThreadPoolExecutor(max_workers=BaosNodeImpl.m_poolSize)   # init thread pool size
        self.m_pool.submit(Multicast.run, self.m_connManager)
        self.m_pool.submit(self.dealpoller)
    
    def init(self):
        self.initialize(g_multicast_addr,None)

    @classmethod
    def getInstance(cls):
        if(cls.m_instance == None):
            lock = Lock()
            lock.acquire()
            if(cls.m_instance == None):
                cls.m_instance = BaosNodeImpl()
            lock.release()
        return cls.m_instance
    
    # create protobuf message
    def createMessage(self,type_name):
        pass

    def runService(self):
        self.m_pool.submit(ZMQRPCServer.run,ZMQRPCServer.getInstance())

    # jsonrpcpp::request_callback f
    def registerService(self,serviceName, f):
        # multicast service
        self.m_connManager.multicastService(serviceName,"tcp://" + str(self.m_localIP) + ":" +  self.m_rpcPort)
        # register call
        ZMQRPCServer.getInstance().registerCall(serviceName, f)

    def callService(self,serviceName, params = None , callback = None):
        endpoints = []
        result = {}
        # if can not find service ,then update 
        if(self.m_connManager.findService(serviceName,endpoints) == False):
            servicecall = ServiceCall()
            servicecall.params = params
            servicecall.callback = callback
            servicecall.servCall = self.callService
            servicecall.serviceName = serviceName
            self.m_connManager.pushServiceCallBack(servicecall)
            self.m_connManager.sendRequest(serviceName, "service")
            return result
        
        # find serve and it's process it's related endpoints
        for endpoint in endpoints:
            # print("service endpoint is ",endpoint)
            # if alread found rpc connection in m_rpcConnection
            if(endpoint in self.m_rpcConnection.keys()):
                if(callback == None):
                    result = self.m_rpcConnection[endpoint].call(serviceName, params)
                else:
                    self.m_rpcConnection[endpoint].async_call(serviceName,params,callback)
            # if not found rpc connection in m_rpcConnection ,then create connection and save it
            else:
                self.m_rpcConnection[endpoint] = ZMQRPCClient()
                self.m_rpcConnection[endpoint].connect(endpoint)
                if(callback == None):
                    result = self.m_rpcConnection[endpoint].call(serviceName, params)
                else:
                    self.m_rpcConnection[endpoint].async_call(serviceName,params,callback)
            return result
        
        
    def publishTopic(self,topic):
        print("topic is ",topic,"  m_topicEndpoint is ",self.m_topicEndpoint)
        self.m_connManager.multicastTopic(topic, self.m_topicEndpoint)
    
    def publish_json(self,topic,json_str):
        sendstr = topic + '-' + str(json_str)
        sendbyte = bytes(sendstr,encoding="ISO-8859-1")
        self.m_pub.send(sendbyte)

        # Publish the data to each client that connects to the node via Websocket and fons to the topic
        pass

    def publish_msg(self,topic,msg):

        class_name = type(msg).__name__     # get object's class name
        msg = msg.SerializeToString()       # SerializeToString

        frame_start = topic + str('-"') 
        frame_end = class_name + str('*,')
        
        # format : topic_name + -classname* + msg
        sendbyte = bytes(frame_start,encoding="utf-8") + len(class_name).to_bytes(1,"big") \
            + bytes(frame_end,encoding="utf-8") + msg
        try:
            self.m_pub.send(sendbyte)
        except zmq.error as e:
            print("zmq publish protobuf msg error : ",e)
        
    def dealRecv(self, msg, flag , index):
        try:
            msg = str(msg,encoding="ISO-8859-1")
            auto_lock = AutoLock(self.m_mutex[self.m_index % BaosNodeImpl.m_poolSize])

            index = msg.find("-")
            topic = msg[0:index]
            content = msg[index+1:]

            if( flag == ProtoType ):
                index = content.find("*")     # find index of real protobuf msg except class name

                content = content[index+2:]     # get msg data
                param = bytes(content,encoding="ISO-8859-1")

                handlers = self.m_msgHandlers[topic]
                for handler in handlers:
                    handler(param)

            elif(flag == JsonType):
                param = json.loads(content)
                handlers = self.m_jsonHandlers[topic]   # handlers list for spectific topic
                for handler in handlers:
                    handler(param)
        except Exception as e:
            print("error is ",e)

    # add alread created and connect socket to list and add socket to zmq poller monitor
    def addSocket(self,socket, flag):
        self.m_poller.register(socket, zmq.POLLIN)
        self.m_socketInsVec.append(socketInfo(socket,flag))

    # zmq poller monitor socket and deal with socket recv
    def dealpoller(self):
        while(1):
            socks = self.m_poller.poll()    # socks [(socket,flag)]  
            # process poller monitor sockets
            for i in range(len(socks)):
                if(socks[i][1] == zmq.POLLIN):  #socket flag == zmq.POLLIN
                    try:
                        auto_lock = AutoLock(self.m_mutex[self.m_index % BaosNodeImpl.m_poolSize])
                        self.m_msgQueue[self.m_index % BaosNodeImpl.m_poolSize] = self.m_socketInsVec[i].socket.recv()

                    except zmq.error as e:
                        print(e)
                        continue
                    self.m_pool.submit(self.dealRecv, self.m_msgQueue[self.m_index % BaosNodeImpl.m_poolSize],
                                        self.m_socketInsVec[i].flag,self.m_index)
                    self.m_index = self.m_index + 1
                    if(self.m_index == BaosNodeImpl.m_poolSize):
                        self.m_index = 0
    
    # create new subcribe socket and add it into connection
    # use addSocket to put socket into poller monitor
    # insert endpoint into connectiondealway which use to check if endpoint is connect or not
    def initNewSub(self,endpoint, topic, flag):
        self.m_connection[endpoint] = self.m_context.socket(zmq.SUB)
        # print("initNewSub connect endpoint is ",endpoint)
        self.m_connection[endpoint].connect(endpoint)
        self.m_connection[endpoint].setsockopt_string(zmq.SUBSCRIBE,topic)

        self.addSocket(self.m_connection[endpoint],flag)

        if(topic not in self.m_connectionDealWay.keys()):
            vec = [0 for i in range(10)]
            self.m_connectionDealWay[endpoint] = vec
    
    def subscribeTopic_Json(self, topic, jsncb):
        endpoints = []
        endpoint = ""

        if( self.m_connManager.findTopic(topic, endpoints) == False ):
            # print("can not find topic")
            jsnCB = JsnCallBack()
            jsnCB.jsnCb = jsncb     # user json callback function
            jsnCB.jsnSub = self.subscribeTopic_Json         # baosnodeimpl.subscribeTopic_Json callback function
            jsnCB.topicName = topic

            self.m_connManager.pushTopic_JsonCallBack(jsnCB)
            self.m_connManager.sendRequest(topic,"topic")
            return

        print("subscribeTopic_Json find endpoints is ",endpoints)
        if(self.m_jsonHandlers.get(topic) == None):
            jsoncb_list = []
            jsoncb_list.append(jsncb)
            self.m_jsonHandlers[topic] = jsoncb_list
        else:
            self.m_jsonHandlers[topic].append(jsncb)

        for endpoint in endpoints:
            if(endpoint not in self.m_connectionJson.keys()):
                self.initNewSub(endpoint, topic, JsonType)
            else:
                self.m_connectionJson[endpoint].setsockopt_string(zmq.SUBSCRIBE, topic)
                if(self.m_connectionDealWay[endpoint][JsonType] == 0):
                    self.addSocket(self.m_connectionJson[endpoint], JsonType)
                    self.m_connectionDealWay[endpoint][JsonType] = 1

    def subscribeTopic_Msg(self, topic, msgcb):
        endpoints = []
        endpoint = ""

        if( self.m_connManager.findTopic(topic, endpoints) == False ):
            # print("can not find topic")
            msgCB = MsgCallBack()
            msgCB.msgCb = msgcb     # user json callback function
            msgCB.msgSub = self.subscribeTopic_Msg         # baosnodeimpl.subscribeTopic_Json callback function
            msgCB.topicName = topic

            self.m_connManager.pushTopic_MsgCallBack(msgCB)
            self.m_connManager.sendRequest(topic,"topic")
            return

        print("subscribeTopic_Msg find endpoints is ",endpoints)
        if(self.m_msgHandlers.get(topic) == None):
            msgcb_list = []
            msgcb_list.append(msgcb)
            self.m_msgHandlers[topic] = msgcb_list
        else:
            self.m_msgHandlers[topic].append(msgcb)

        for endpoint in endpoints:
            if(endpoint not in self.m_connection.keys()):
                self.initNewSub(endpoint, topic, ProtoType)
            else:
                self.m_connection[endpoint].setsockopt_string(zmq.SUBSCRIBE, topic)
                if(self.m_connectionDealWay[endpoint][ProtoType] == 0):
                    self.addSocket(self.m_connection[endpoint], ProtoType)
                    self.m_connectionDealWay[endpoint][ProtoType] = 1
    
    # get port automaticly allocated by os
    def get_free_port(self):  
        sock = socket.socket()
        sock.bind(('', 0))
        ip, port = sock.getsockname()
        sock.close()
        return port

    def unSubscribeTopic(self,topic):
        endpoints = []

        self.m_connManager.findTopic(topic,endpoints)

        # clear json data type connection and handlers
        for endpoint in endpoints:
            if( endpoint in self.m_connectionJson.keys() ):
                # cancel subscribe
                for i in range(len(self.m_jsonHandlers[topic])):
                    self.m_connectionJson[endpoint].setsockopt_string(zmq.UNSUBSCRIBE,topic)
                del self.m_jsonHandlers[topic]
                print("unsubscribe ",topic)

        # clear protobuf data type connection and handlers
            if(endpoint in self.m_connection.end()):
                for i in range(len(self.m_msgHandlers[topic])):
                    self.m_connection[endpoint].setsockopt_string(zmq.UNSUBSCRIBE,topic)
                del self.m_msgHandlers[topic]
                priint("unsubscribe ",topic)